/*

  bprom
  for use with Dr. Boehm u-PROM and u-PROM 2001 
  (EPROM burner with serial interface)

  Nils Eilers <nils.eilers@gmx.de>

  first edit: 2007-11-14
  last edit : 2008-11-18

*/

#include <stdio.h>   	/* Standard input/output definitions */
#include <string.h>  	/* String function definitions */
#include <ctype.h>	/* toupper */
#include <unistd.h>  	/* UNIX standard function definitions */
#include <fcntl.h>   	/* File control definitions */
#include <errno.h>   	/* Error number definitions */
#include <termios.h> 	/* POSIX terminal control definitions */
#include <stdlib.h>	/* exit() */
#include <stdarg.h>	/* variable arguments */
#include <sys/stat.h>	/* file permissions */
#include <time.h>	/* nanosleep */
#include "cmdline.h"	/* Command line parameters */
#include "bprom.h"	/* Global definitions */
#include "prom2001.h"	/* Programmer model u-prom 2001 */
#include "uprom.h"	/* Programmer model u-prom */
#include "config.h"	/* Configuration, defaults */
#include "bootloader.h"	/* bprom as bootloader */

int varinit(void)
{
  verbosity=VB_ERR;
  ictyp=mc=dc=nr=from=size=erraddr=0;
  strcpy(devicename, DEFAULT_SERIAL_PORT);
  Hardware_Handshake = DEFAULT_HW_HANDSHAKE;
  Data_Bits=DEFAULT_DATA_BITS;
  Stop_Bits=DEFAULT_STOP_BITS;
  Parity=DEFAULT_PARITY;
  Baud_Rate=DEFAULT_BAUD_RATE_UPROM2001;

  buffer=calloc(BUFSIZE, sizeof(BYTE));
  if(buffer==NULL) {
    perror("can't allocate memory");
    return(3);
  }

  return(0);
}

/* logging and verbose dependand output */

void lvprintf(int v, const char *fmt, ...)
{
  va_list arg_ptr;
  int l;
  
  l=v; if(v>=0x1000) return;	/* log only? */
  
  if(verbosity<v) return;
  
  va_start(arg_ptr, fmt);
  vfprintf(stdout, fmt, arg_ptr);
  va_end(arg_ptr);
}

/* status line */
int statln(long int pos, long int total)
{
  static long int oldpos;
  int width=64, p, i;
  
  if(verbosity!=VB_INFO) return 0;
  
  if(pos==(-1)) {
    printf("\n"); fflush(stdout);
    return 0;
  }

  if(((pos-oldpos)>=(total/width)) || (pos==0) || (pos==total)) {
    p=((float)pos/(float)total*width);
    if(p) for(i=0; i<p; i++) putchar('=');
    for(i=0; i<(width-p); i++) putchar('-');
    putchar(' ');

    oldpos=pos;
    printf("%04lX/%04lX%c", pos, total, CR);
    fflush(stdout);
  }

/*  
  else if(kbhit()) if(getch()==ESC) {
    printf("\n"); fflush(stdout);
    return 1;
  }
*/  
  return 0;
}

int endstatln(long from, long total)
{
  int exitcode;
  
  if(verbosity<VB_INFO) return 0;
  exitcode=statln(from, total);
  printf("\n");
  return exitcode;
}

/* open serial port */
/* returns file descriptor on success or -1 on error */
                                           
int open_port(int send_break)
{
  int fd; /* File descriptor for the port */

  fd = open(devicename, O_RDWR | O_NOCTTY | O_NDELAY);
  if (fd == -1)
  {
    fprintf(stderr, "%s: ", devicename); 
    perror(devicename);			/* could not open the port */
    exit(3);
  }
  else fcntl(fd, F_SETFL, 0);		/* wait for data on reading */

  /* get the current options */
  if(tcgetattr(fd, &sio)) {
    fprintf(stderr, "%s: ", devicename); 
    perror("unable to get current status of serial port");
  };

  switch (Baud_Rate)
  {
    case 38400:
      BAUD = B38400;
      break;
    case 19200:
      BAUD  = B19200;
      break;
    default:
    case 9600:
      BAUD  = B9600;
      break;
    case 4800:
      BAUD  = B4800;
      break;
    case 2400:
      BAUD  = B2400;
      break;
    case 1800:
      BAUD  = B1800;
      break;
    case 1200:
      BAUD  = B1200;
      break;
    case 600:
      BAUD  = B600;
      break;
    case 300:
      BAUD  = B300;
      break;
    case 200:
      BAUD  = B200;
      break;
    case 150:
      BAUD  = B150;
      break;
    case 134:
      BAUD  = B134;
      break;
    case 110:
      BAUD  = B110;
      break;
    case 75:
      BAUD  = B75;
      break;
    case 50:
      BAUD  = B50;
      break;
  } /* end of switch baud_rate */

  switch (Data_Bits)
  {
    case 8:
    default:
      DATABITS = CS8;
      break;
    case 7:
      DATABITS = CS7;
      break;
    case 6:
      DATABITS = CS6;
      break;
    case 5:
      DATABITS = CS5;
      break;
  }  /* end of switch data_bits */

  switch (Stop_Bits)
  {
    case 1:
      STOPBITS = 0;
      break;
    default:
    case 2:
      STOPBITS = CSTOPB;
      break;
  }  /* end of switch stop bits */

  switch (Parity)
  {
    case 0:
    default:                       /* none */
      PARITYON = 0;
      PARITY = 0;
      break;
    case 1:                        /* odd */
      PARITYON = PARENB;
      PARITY = PARODD;
      break;
    case 2:                        /* even */
      PARITYON = PARENB;
      PARITY = 0;
      break;
  }  /* end of switch parity */

  /* Set baud rates (new style) */
  if(cfsetispeed(&sio, BAUD)) {
    perror("set input baud rate");
  }
  if(cfsetospeed(&sio, BAUD)) {
    perror("set output baud rate");
  }
  /* set new port settings for raw data processing */
  sio.c_cflag = BAUD | DATABITS | STOPBITS | PARITYON | PARITY | CLOCAL | CREAD;

  if(Hardware_Handshake) {
    /* ENABLE RTS/CTS */
    sio.c_cflag |= CRTSCTS; 			/* aka CNEW_RTSCTS 	*/
  } else 
  {
    /* DISABLE RTS/CTS */
    sio.c_cflag &= ~CRTSCTS; 
  }

  sio.c_iflag = IGNPAR;				/* ignore parity errors	*/
  sio.c_iflag &= ~(IXON | IXOFF | IXANY); 	/* hardware flow control*/
  sio.c_oflag = 0;				/* raw output 		*/
  
  sio.c_lflag=0; 				/* raw mode, no echo	*/
  sio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);


  /* Timeouts are ignored in canonical input mode or when the 		*/
  /*  NDELAY option is set on the file via open or fcntl 		*/
  sio.c_cc[VMIN]=1;	/* Minimum number of characters to read 	*/
  sio.c_cc[VTIME]=20;	/* Time to wait for data (tenths of seconds)	*/

  if(tcflush(fd, TCIOFLUSH)) {	/* flush read and write buffer */
    perror("flush serial read and write buffer");
  }
  lvprintf(VB_DBG, "Serial Buffer flushed\n");
  if(tcsetattr(fd,TCSANOW,&sio)) {
    perror("set serial attributes");
  }
  lvprintf(VB_DBG, "Serial port attributes set\n");

  if(send_break) {
    lvprintf(VB_DBG, "Sending BREAK to reset programmer, please wait...\n");
    if(tcsendbreak(fd, 5)) {
      perror("send break");
    }
    sleep(1);	/* send break to reset programmer */
    lvprintf(VB_DBG, "Break sent\n");
  }
  
  lvprintf(VB_DBG, "Serial port opened.\n");

  return (fd);
}

int close_port(int send_break)
{
  int error;
  
  error=0;
  if(send_break) {
    if(tcsendbreak(fd, 5)) {
      perror("send break");
      error=error | 1;
    }
    sleep(5);	/* send break to reset programmer */
  }
  
  if(tcflush(fd, TCIOFLUSH)) {	/* flush read and write buffer */
    perror("flush serial read and write buffer");
    error=error | 2;
  }
  switch(close(fd)) {				/* close serial port */
    case 0: 
      lvprintf(VB_DBG, "Serial port closed.\n");
      break;
    case 1:
      perror("close serial port");
      error=error | 4;
      break;
    default:
      perror("close serial port/unknown error");
      error=error | 8;
  }
  return error;
}

void long2hml(long v, HML *byte)
{
  byte->h=(v & 0xFF0000) >> 16;
  byte->m=(v & 0x00FF00) >> 8;
  byte->l=v & 0x0000FF;
}

long hml2long(HML *bytes)
{
  return(bytes->l | (bytes->m << 8) | (bytes->h << 16));
}


void errhdl(int errcde)
{
  if(errcde==0) return;
  if(errcde<0) {
    lvprintf(VB_ERR, "Error %d: ", errcde);
    switch(errcde) {
      case -1:
        lvprintf(VB_ERR, "CRC error\n");
        break;
      default:
        lvprintf(VB_ERR, "Unknown internal error %d\n", errcde);
        break;
    } /* switch(errcde) */
  } else 
  {
    if((errcde>0x80) && (errcde<0x98)) {
      lvprintf(VB_ERR, "Error %02X: %s\n", errcde, errmsg[errcde-0x81]);
    } else {
      lvprintf(VB_ERR, "unknown error %02X\n", errcde);
    }
  }
}    

void show_buffer(long from, long size)
{
  long i, x;
  
  x=0;
  for(i=from;  i<=size; i++) {
    if(x == DUMP_WIDTH) {
      for(x = i - DUMP_WIDTH; x < i; x++) 
        if(buffer[x]>31 && buffer[x]<128) lvprintf(VB_DBG, "%c", buffer[x]);
        else lvprintf(VB_DBG, ".");
      x=0;
    }
    if(i==size) break;
    if(!x) lvprintf(VB_DBG, "\n%04X ", i);
    lvprintf(VB_DBG, "%02X ", buffer[i] & 0xFF);
    x++; if(x==4) lvprintf(VB_DBG, ": ");
  }
  lvprintf(VB_DBG, "\n");
  lvprintf(VB_DBG, "Checksum in buffer: %02X\n", buffer[size]);
}  

int save_buffer(long size, long offset, char *filename)
{
  long i;
  FILE *out;
  
  lvprintf(VB_DBG, "save_buffer (size= %lXh offset=%lXh filename=%s\n", size, offset, filename);
  
  if((out=fopen(filename, "wb"))==NULL) {
    fprintf(stderr, "%s: ", filename); perror("unable to open file");
    return 3;
  }
  
  i=fwrite(buffer+offset, sizeof(BYTE), size, out);
  if(i<size) {
    fprintf(stderr, "%s: ", filename); perror("cannot write data to file");
    return 3;
  }
  return(fclose(out) ? 3 : 0);
}  

int load_buffer(long size, long offset, char *filename)
{
  size_t rd;
  FILE *in;
  
  if((in=fopen(filename, "rb"))==NULL) {
    fprintf(stderr, "%s: ", filename); 
    perror("unable to open file");
    return 3;
  }
  
  rd=fread(buffer+offset, sizeof(BYTE), size, in);
  if(rd<size) {
    if(!rd) {
      fprintf(stderr, "%s: ", filename); 
      perror("cannot read file");
    }
    else {
      lvprintf(VB_ERR, "Should read %ld bytes but could only read %ld.\n", size, rd);
      perror("");
    }
    return 3;
  }
  return(fclose(in) ? 3 : 0);
}  

int calc_crc(long from, long size)
{
  int crc=0; long i;

  for(i=from; i<(from+size); i++) crc=(crc+buffer[i]) & 0xFF;
  lvprintf(VB_DBG, "sum: %02X\n", crc);
  crc=(0xFF - crc) & 0xFF;
  lvprintf(VB_DBG, "CRC: %02X\n", crc);
  return crc;
}

void show_cmdbuffer(char *cmdbuffer, int size)
{
  int i;

  lvprintf(VB_DBG, "Command string:\n");
  for(i=0; i<size; i++) {
    lvprintf(VB_DBG, "%02X ", cmdbuffer[i] & 0xFF);
  } lvprintf(VB_DBG, "\n");
}

ssize_t write_delayed(int fildes, const void *buf, size_t nbyte, long delay)
{
  size_t i;
  struct timespec t;
  
  t.tv_sec=0;
  t.tv_nsec=delay;

  for(i=0; i<nbyte; i++) {
    if(write(fildes, buf, 1)<1) {
      /* oops, we could not write it! */
      return i;
    }
    nanosleep(&t, NULL);  
  }
  return nbyte;
}

int send_cmdbuffer(char *cmdbuffer, int size)
{
  lvprintf(VB_DBG, "Sending command string...\n");
  if(delay_serial) return write_delayed(fd, cmdbuffer, size, delay_serial);
  return write(fd, cmdbuffer, size);
}

void dump_args(void)
{
  lvprintf(VB_DBG, "Command line options:\n\n");
  lvprintf(VB_DBG, "Serial port: %s\n", devicename); 
  lvprintf(VB_DBG, Hardware_Handshake \
                          ? "Use RTS/CTS\n":"No Hardware handshake\n");
  lvprintf(VB_DBG, "Baudrate: %ld\n", Baud_Rate);
  lvprintf(VB_DBG, "Verbosity: %d\n", verbosity);
  lvprintf(VB_DBG, "ictyp: %02Xh\n", ictyp);
  lvprintf(VB_DBG, "from: %lXh size: %lXh\n", from, size); 
}  


int main(int argc, char**argv)
{
  int programmer=DEFAULT_PROMMER;
  int errcde=0;
  int exitcode=0;
  struct gengetopt_args_info args_info;
  
  if(varinit()) exit(3);
  
  umask(0 | S_IWGRP | S_IWOTH);		/* default permissions rw-r--r-- */
  
  /* parse command line options */  
  if(cmdline_parser(argc, argv, &args_info) != 0) exit(3);
  if(argc<2) {				/* any arguments or options? */
    cmdline_parser_print_help();	/* no: give usage */
    exit(3);
  }
  
  
				  	/* Serial port settings */
  if(args_info.serial_port_given) strncpy(devicename, args_info.serial_port_arg, 80);
  if(args_info.no_handshake_given) Hardware_Handshake=FALSE;
  delay_serial=0; if(args_info.delay_given) delay_serial=args_info.delay_arg;
  
  Baud_Rate=DEFAULT_BAUD_RATE_UPROM2001;
  if(args_info.baud_rate_given) Baud_Rate=atol(args_info.baud_rate_arg);
  
  if(args_info.verbose_given) verbosity=VB_INFO;
  if(args_info.debug_given) verbosity=VB_DBG;

  
  from=0; /* if(args_info.from_given) from=args_info.from_arg; */
  offset=0; /* if(args_info.offset_given) offset=args_info.offset_arg; */
  if(args_info.size_given) size=args_info.size_arg;
  
  
  if(args_info.programmer_given) {
    programmer=DEFAULT_PROMMER;
    if(!strncmp(args_info.programmer_arg, "uprom", 32)) programmer=UPROM; 
    else if(!strncmp(args_info.programmer_arg, "uprom2001", 32)) programmer=UPROM2001; 
    else if(!strncmp(args_info.programmer_arg, "bootloader", 32)) programmer=BOOTLOADER; 
    else {
      lvprintf(VB_ERR, "Unknown programmer %ls\n", args_info.programmer_arg);
      return(3);
    }
  }

  switch(programmer) {
    case UPROM:
      errcde=uprom(&args_info, argc); break;
    case UPROM2001:
      errcde=uprom2001(&args_info, argc); break;
    case BOOTLOADER:
      errcde=bootloader(&args_info, argc); break;
  }    
  
  free(buffer);	buffer=NULL;		/* free memory */
  /* exitcode:
     0 = everything okay!
     1 = operation completed, but eprom not blank or compare found differences
     2 = failed, normal exit
     3 = fatal errors, aborted
  */
  return(errcde ? 2 : exitcode);
}
