/*

  prom2001 - for use with Boehm y-PROM 2001 
  (EPROM burner with serial interface)

  Nils Eilers <nils.eilers@gmx.de>

  first edit: 2007-11-14
  last edit : 2007-11-24
  last edit : 2008-02-16
  	- more intuitive command line parameters- 
  last edit : 2008-03-25
  	- split into modules
  last edit : 2008-11-18
        - open_port(TRUE) send break to reset programmer
*/

#include <stdio.h>   	/* Standard input/output definitions */
#include <string.h>  	/* String function definitions */
#include <unistd.h>  	/* UNIX standard function definitions */
#include <fcntl.h>   	/* File control definitions */
#include <errno.h>   	/* Error number definitions */
#include <termios.h> 	/* POSIX terminal control definitions */
#include <stdlib.h>	/* exit() */
#include <stdarg.h>	/* variable arguments */
#include <sys/stat.h>	/* file permissions */

#include <cmdline.h>
#include <bprom.h>
#include <prom2001.h>

char errmsg[23][32] = { 
/*0....:....1....:....2....:....3....:....4....:....5....:....6....:...7 */
  /* 0x81 */	"Undefined command", 
  /* 0x82 */	"Undefined chip", 
  /* 0x83 */	"Wrong chip nr.",
  /* 0x84 */	"Unknown or no identifier", 
  /* 0x85 */	"(E)EPROM is not empty", 
  /* 0x86 */	"Contents mismatch",
  /* 0x87 */	"Could not erase EEPROM", 
  /* 0x88 */	"Start address beyond limits", 
  /* 0x89 */	"End address beyond limits", 
  /* 0x8A */	"-- 0x8A --",
  /* 0x8B */	"-- 0x8B --",
  /* 0x8C */	"-- 0x8C --",
  /* 0x8D */	"-- 0x8D --",
  /* 0x8E */	"-- 0x8E --",
  /* 0x8F */	"-- 0x8F --",
  /* 0x90 */	"Insufficient buffer size", 
  /* 0x91 */	"CRC error", 
  /* 0x92 */	"Unable to reprogram", 
  /* 0x93 */	"Unable to program", 
  /* 0x94 */	"Verify error",
  /* 0x95 */	"No data", 
  /* 0x96 */	"Cannot erase this EEPROM", 
  /* 0x97 */	"Serial communication error" };


int uprom2001(struct gengetopt_args_info *args_info, int argc)
{
  int cmd, errcde, exitcode;

  lvprintf(VB_DBG, "programmer model: uprom2001\n");

  ictyp= -1; 
  
  /* chip type and programming mode */

  if(args_info->type_given) {
    if(!strncasecmp(args_info->type_arg, "48T08", 10)) {
      ictyp=32;
      size=0x1ff8; from=0x2000; offset=0x2000;
    }
    if(ictyp<0) {
      lvprintf(VB_ERR, "prom2001: Chip %s not recognized or not supported.\n", args_info->type_arg);
      return(3);
    }
  } 

  if(args_info->mode_given) {
    if(!strncasecmp(args_info->mode_arg, "0x", 2)) {
      if(sscanf(args_info->mode_arg, "%X", &ictyp)!=1) {
        lvprintf(VB_ERR, "invalid argument %s for -m\n", args_info->mode_arg);
        return(3);
      }
    } else {
      if(sscanf(args_info->mode_arg, "%d", &ictyp)!=1) {
        lvprintf(VB_ERR, "invalid argument %s for -m\n", args_info->mode_arg);
        return(3);
      }
    }
    if(!args_info->size_given) size=sizefromictyp(ictyp);
  }    

  if(args_info->size_given) {
    size=args_info->size_arg;
  }
  if(args_info->from_given) {
    from=args_info->from_arg;
  }
  
  if(ictyp<0) {
    lvprintf(VB_ERR, "Please specify chip type and/or programming mode!\n");
    return(3);
  }
  
  dump_args();
  fd=open_port(TRUE);
  cmd= -1;		
  if(args_info->auto_given) {				/* autodetect chip */
    cmd=CMD_GET_ID; if(verbosity<VB_INFO && argc==2) verbosity=VB_INFO;
    if((exitcode=action(cmd))) return(exitcode); 
  }

  if(args_info->erase_given) {				/* erase EEPROM */
    cmd=CMD_ERASE;
    if((exitcode=action(cmd))) return(exitcode); 
  } 
  
  if(args_info->blank_given) {				/* blank test */
    cmd=CMD_BLANK_COPY;
    if((exitcode=action(cmd))) return(exitcode); 
  } 
  
  if(args_info->write_given) {				/* write chip */
    if(args_info->master_socket_given) {
      lvprintf(VB_ERR, "Master socket is for reading only! Use copy socket.\n");
      return(3);
    }
    if(args_info->blank_given) sleep(1);
    errcde=load_buffer(size, offset, args_info->write_arg);
    if(errcde) return 3;
    cmd=CMD_PRG_COPY;
    errcde=action(cmd);
    if(errcde) return errcde;
  }  
  
  if(args_info->read_given) {				/* read chip */
    cmd=CMD_READ_COPY;
    if(args_info->master_socket_given) cmd=CMD_READ_MASTER;
    errcde=action(cmd);
    if(errcde) return errcde;
    errcde=exitcode=save_buffer(size, offset, args_info->read_arg);
    if(errcde) return errcde;
  }  
  
  if(args_info->compare_given) {	/* compare master == copy */
    cmd=CMD_COPY_EQU_MASTER;
    action(cmd);
  }
  
  return close_port(TRUE); 
}

int read_eprom(int cmd, int ictyp, long from, long size)
{
  int rd, crc, crc_soll, errcde;
  long sum; 
  HML hml_from; HML hml_size;
  char cmdbuffer[16];
  
  lvprintf(VB_DBG, "read_eprom: ictyp=%02X from=%lX size=%lX\n", \
                                            ictyp, from, size);
  long2hml(from, &hml_from); long2hml(size, &hml_size);  
  
  cmdbuffer[0]=cmd;
  cmdbuffer[1]=ictyp;
  cmdbuffer[2]=hml_from.h;
  cmdbuffer[3]=hml_from.m;
  cmdbuffer[4]=hml_from.l;
  cmdbuffer[5]=hml_size.h;
  cmdbuffer[6]=hml_size.m;
  cmdbuffer[7]=hml_size.l;
  show_cmdbuffer(cmdbuffer, 8);
  send_cmdbuffer(cmdbuffer, 8);  

  lvprintf(VB_DBG, "Waiting for result...\n");
  do {
    rd=read(fd, buffer, 1);
  } while(!rd);

  errcde=buffer[0];
  lvprintf(VB_DBG, "Result: %X\n", errcde);
    
  if(errcde) {
    return(errcde);    
  } else
  {
    lvprintf(VB_DBG, "Reading data...\n");
    sum=0;
    while(sum<(size+1)) {
      statln(sum, size);
      rd=read(fd, buffer+from+sum, size+1-sum);
      if(rd<0) {
        perror("reading serial data");
        break;
      }
      sum+=rd;
      lvprintf(VB_DBG, "%d bytes received, total %ld of %ld+1.\n", rd, sum, size);
    }
    endstatln(size, size); 

    show_buffer(from, size);

    crc_soll=buffer[from+size] & 0xFF;
    lvprintf(VB_DBG, "CRC (prommer): %02X\n", crc_soll);

    crc=calc_crc(from, size);

    lvprintf(VB_DBG, (crc==crc_soll) ? "CRC OK\n" : "CRC ERROR\n");
    if(crc==crc_soll) return 0;
    else return -1;
  } /* if(errcde) */
}

int copy_equ_master(int ictyp)
{
  int rd, errcde, v;
  char cmdbuffer[16];
  char buffer[2];
  
  lvprintf(VB_DBG, "copy_equ_master: ictyp=%02X", ictyp);
  cmdbuffer[0]=CMD_COPY_EQU_MASTER;
  cmdbuffer[1]=ictyp;
  show_cmdbuffer(cmdbuffer, 2);
  send_cmdbuffer(cmdbuffer, 2);

  do {
    rd=read(fd, buffer, 1);
  } while(!rd);
  errcde=buffer[0] & 0xFF;
  lvprintf(VB_DBG, "er: %X\n", errcde);

  do {
    rd=read(fd, buffer, 1);
  } while(!rd);
  v=buffer[0] & 0xFF;
  lvprintf(VB_DBG, "vv: %X\n", v);
  
  return(errcde ? errcde : v);
}

int blank_copy(int ictyp)
{
  int rd, errcde, bc;
  char cmdbuffer[16];
  char buffer[2];
  
  lvprintf(VB_DBG, "blank_copy: ictyp=%02X\n", ictyp);
  
  cmdbuffer[0]=CMD_BLANK_COPY;
  cmdbuffer[1]=ictyp;
  show_cmdbuffer(cmdbuffer, 2);
  send_cmdbuffer(cmdbuffer, 2);

  do {
    rd=read(fd, buffer, 1);
  } while(!rd);
  errcde=buffer[0] & 0xFF;
  lvprintf(VB_DBG, "er: %X\n", errcde);

  do {
    rd=read(fd, buffer, 1);
  } while(!rd);
  bc=buffer[0] & 0xFF;
  lvprintf(VB_DBG, "vv: %X\n", bc);
  
  return(errcde ? errcde : bc);
}

int get_id(void)
{
  int rd, errcde, sum;
  char cmdbuffer[16];
  char buffer[4];
  
  lvprintf(VB_DBG, "get_id\n");
  
  cmdbuffer[0]=CMD_GET_ID;
  show_cmdbuffer(cmdbuffer, 1);
  send_cmdbuffer(cmdbuffer, 1);


  sum=0;
  while(sum<4) {
    rd=read(fd, buffer+sum, 4-sum);
    sum+=rd;
  }
  errcde=buffer[0] & 0xFF;
  mc=buffer[1] & 0xFF;
  dc=buffer[2] & 0xFF;
  nr=buffer[3] & 0xFF;
  lvprintf(VB_DBG, "er: %X\n", errcde);

  return(errcde);
}

int eeprom_erase(int ictyp)
{
  int rd, errcde, sum, ll;
  char cmdbuffer[16];
  char buffer[4];
  
  lvprintf(VB_DBG, "eeprom_erase: ictyp=%02X\n", ictyp);
  cmdbuffer[0]=CMD_ERASE;
  cmdbuffer[1]=ictyp;
  show_cmdbuffer(cmdbuffer, 2);
  send_cmdbuffer(cmdbuffer, 2);

  sum=0;
  while(sum<2) {
    rd=read(fd, buffer+sum, 2-sum);
    sum+=rd;
  }
  errcde=buffer[0] & 0xFF;
  ll=buffer[1] & 0xFF;
  lvprintf(VB_DBG, "er: %02X\n", errcde);
  lvprintf(VB_DBG, "ll: %02X\n", ll);
  if((errcde==0) & (ll!=0)) errcde=ll;
  if(!errcde) lvprintf(VB_INFO, "EEPROM erased.\n");

  return(errcde);
}

int prg_1buffer(int ictyp, long from, long size)
{
  int rd, wr, crc, errcde;
  long sum; 
  HML hml_from; HML hml_size; HML ea;
  char cmdbuffer[16];
  char rcbuffer[4];
  
  lvprintf(VB_DBG, "prg_1buffer: ictyp=%02Xh from=%lXh size=%lXh\n", ictyp, from, size);
  
  long2hml(from, &hml_from); long2hml(size, &hml_size);  
  erraddr=0;
  
  crc=calc_crc(from, size);
  cmdbuffer[8]=crc;
  
  cmdbuffer[0]=CMD_PRG_COPY;
  cmdbuffer[1]=ictyp;
  cmdbuffer[2]=hml_from.h;
  cmdbuffer[3]=hml_from.m;
  cmdbuffer[4]=hml_from.l;
  cmdbuffer[5]=hml_size.h;
  cmdbuffer[6]=hml_size.m;
  cmdbuffer[7]=hml_size.l;
  show_cmdbuffer(cmdbuffer, 8);
  send_cmdbuffer(cmdbuffer, 8);  
  
  wr=write(fd, buffer+from, size);		/* programm data */
  
  write(fd, cmdbuffer+8, 1);  			/* crc */

  lvprintf(VB_DBG, "Waiting for result...\n");
  do {
    rd=read(fd, rcbuffer, 1);
  } while(!rd);

  errcde=rcbuffer[0] & 0xFF;
  lvprintf(VB_DBG, "Result: %X\n", errcde);
  if(errcde) return(errcde);
  
  lvprintf(VB_DBG, "Waiting for result...\n");
  do {
    rd=read(fd, rcbuffer, 1);
  } while(!rd);

  errcde=rcbuffer[0] & 0xFF;
  lvprintf(VB_DBG, "Result: %X\n", errcde);
    
  if(errcde) {				/* failed */
    switch(errcde) {
      case 0x92:
      case 0x93:
      case 0x94:
                                        /* reading address */
        sum=0;				/* where error occured */
        while(sum<3) {
          rd=read(fd, rcbuffer+sum, 3-sum);
          sum+=rd;
        }
        ea.h=rcbuffer[0];
        ea.m=rcbuffer[1];
        ea.l=rcbuffer[2];
        erraddr=hml2long(&ea);
        break;
        
      default:
        break;
    } /* switch(errcde) */
  } /* if (errcde) */
  return(errcde);    
}

int prg_copy(int ictyp, long from, long size)
{
  int errcde;			/* error code */
  long current_size; 		/* bytes to be written this time */
  long sum;			/* bytes already written */
  long total;			/* bytes yet to be written */
  
  lvprintf(VB_DBG, "prg_copy: ictyp=%02X from=%lX size=%lX\n", \
                                              ictyp, from, size);

  sum=0; total=size;  
  do {
    current_size=total;
    if(current_size>PROMBUFSIZE) {
      total-=PROMBUFSIZE;
      current_size=PROMBUFSIZE;
    }
    lvprintf(LOG_INFO, \
    "Programming %lXh bytes starting at %lXh\n", current_size, from+sum);
    statln(sum, size);
    errcde=prg_1buffer(ictyp, from+sum, current_size);
    if(errcde) {
      statln(-1, 0);
      return errcde;
    }
    sum+=current_size;
  } while(sum<size);    
  endstatln(size, size);
  return errcde;
}

int action(int cmd)
{
  int errcde, exitcode;
  
  lvprintf(VB_DBG, "action %d: ", cmd);
  errcde=exitcode=0;
  

  switch(cmd) {
    case CMD_READ_MASTER:
    case CMD_READ_COPY:
      errcde=exitcode=read_eprom(cmd, ictyp, from, size);
      break;
    case CMD_COPY_EQU_MASTER:
      errcde=copy_equ_master(ictyp);
      if(errcde) {
        if(errcde==0x86) {
          lvprintf(VB_ERR, "Different data.\n");
          errcde=0; exitcode=1;
        } 
      } else
      {
        lvprintf(VB_INFO, "Equal data.\n");
      }
      break;
      
    case CMD_BLANK_COPY:
      errcde=blank_copy(ictyp);
      if(errcde) {
        if(errcde==0x85) {
          lvprintf(VB_ERR, "Not blank.\n");
          errcde=0; exitcode=1;
        } 
      } else {
        lvprintf(VB_INFO, "Blankcheck passed.\n");
      } 
      break;
    
    case CMD_GET_ID:
      errcde=get_id();
      if(errcde==0x84 && ((mc & dc & nr)==0xFF)) {
        /* lvprintf(VB_ERR, "No identifier\n"); */
        exitcode=1;
      } else {
        lvprintf(VB_INFO, "Manufacturer: %02hX\n", mc);
        lvprintf(VB_INFO, "Device code : %02hX\n", dc);
        lvprintf(VB_INFO, "Internal type nr: %02Xh\n", nr);
        if(nr==0x90) {
          lvprintf(VB_INFO, "Unknown chip\n");
        } else {
          lvprintf(VB_INFO, "Known chip\n");
          ictyp=nr;
        }
      }
      break;
    case CMD_ERASE:
      errcde=exitcode=eeprom_erase(ictyp);
      break;
      
    case CMD_PRG_COPY:
      errcde=exitcode=prg_copy(ictyp, from, size);
      if(exitcode) lvprintf(VB_ERR, "At %lXh: ", erraddr);
      break;

    default:
      lvprintf(VB_ERR, "Missing command (see help/actions)\n");
      return(3);
  }
  if(errcde) errhdl(errcde);
  return(exitcode);
}

long sizefromictyp(int ictyp)
{
  if((ictyp>= 0) && (ictyp<=62))   return(2048 << (ictyp/7));
  if((ictyp>=84) && (ictyp<=128)) return(2048 << ((ictyp-84)/5)); 
  return -1;
}

