/* uprom.c
   for use with Dr. Boehm u-PROM
   first edit 2008-03-25
   last  edit 2008-04-17
*/

#include <stdio.h>   	/* Standard input/output definitions */
#include <string.h>  	/* String function definitions */
#include <unistd.h>  	/* UNIX standard function definitions */
#include <stdlib.h>	/* exit() */
#include <errno.h>   	/* Error number definitions */
#include <fcntl.h>   	/* File control definitions */
#include <termios.h> 	/* POSIX terminal control definitions */
#include <stdarg.h>	/* variable arguments */
#include <sys/stat.h>	/* file permissions */

#include <config.h>
#include <cmdline.h>
#include <bprom.h>
#include <uprom.h>

int uprom(struct gengetopt_args_info *args_info, int argc)
{
  int exitcode, errcde, cmd;
  
  exitcode=errcde=cmd=0;

  lvprintf(VB_DBG, "programmer model: uprom\n");
  
  uprom_init_errmsg();
  if(!args_info->baud_rate_given) {
    Baud_Rate=DEFAULT_BAUD_RATE_UPROM;
    lvprintf(VB_DBG, "Baud rate set to %d = default for uprom programmer\n", Baud_Rate);
  }

  if(args_info->auto_given) {
    lvprintf(VB_ERR, "This programmer is unable to autodetect chips!\n");
    return(3);
  }
  if(args_info->write_given) {			/* write chip */
    if(args_info->master_socket_given) {
      lvprintf(VB_ERR, "Master socket is for reading only! Use copy socket.\n");
      return(3);
    }
  }

  /* chip type */
  if(!args_info->type_given) {
    lvprintf(VB_ERR, "Use -t <type> to specify the chip, e.g. -t 27256\n");
    return(3);
  }
  ictyp= -1;
  if(!strncmp(args_info->type_arg, "2716", 10)) ictyp=0;
  if(!strncmp(args_info->type_arg, "2816", 10)) ictyp=2816;
  if(!strncmp(args_info->type_arg, "2732", 10)) ictyp=1;
  if(!strncasecmp(args_info->type_arg, "2732A",10)) ictyp=1;
  if(!strncmp(args_info->type_arg, "2764", 10)) ictyp=2;
  if(!strncmp(args_info->type_arg, "27128",10)) ictyp=3;
  if(!strncmp(args_info->type_arg, "27256",10)) ictyp=4;
  if(!strncmp(args_info->type_arg, "2532", 10)) ictyp=5;
  if(ictyp<0) {
    lvprintf(VB_ERR, "uprom: Chip %s not recognized or not supported.\n", args_info->type_arg);
    return(3);
  } 

  if(!args_info->mode_given) {
    uprom_prog_mode=DEFAULT_UPROM_PROG_MODE;
  } else
  {
    if(!strncasecmp(args_info->mode_arg, "STD", 8)) {
      uprom_prog_mode=PROG_MODE_STD;
    } else
    if(!strncasecmp(args_info->mode_arg, "SLOW", 8)) {
      uprom_prog_mode=PROG_MODE_STD;
    } else
    if(!strncasecmp(args_info->mode_arg, "FAST", 8)) {
      uprom_prog_mode=PROG_MODE_FAST;
    } else
    if(!strncasecmp(args_info->mode_arg, "FST", 8)) {
      uprom_prog_mode=PROG_MODE_FAST;
    } else {
      lvprintf(VB_ERR, "Illegal value %s for -m\n", args_info->mode_arg);
      return(3);
    }
  }    


  if(args_info->size_given) 
  {
    size=args_info->size_arg;
    lvprintf(VB_DBG, "Size from command line: %ld\n", size);
  } else
  {
    size=uprom_sizefromictyp(ictyp);
    lvprintf(VB_DBG, "Size assumed by type: %ld\n", size);
  }
  
  from=0; if(args_info->from_given) from=args_info->from_arg;
  offset=0; if(args_info->offset_given) offset=args_info->offset_arg;
  
  if(args_info->erase_given) {
    if(ictyp!=2816) {
      lvprintf(VB_ERR, "This programmer can only erase 2816 EEPROMs.\n");
      return(3);
    } 
  }
  
  if(args_info->compare_given) {
    lvprintf(VB_ERR, "This function is not implemented for this programmer. Use the VERIFY key on the programmer.\n");
    return(3); 
  }

  dump_args();
  /* ***** NOW OPEN PORT ***** */
  fd=open_port(FALSE);			/* dont send BREAK or prgr hangs */

  if(args_info->erase_given) {		/* erase EEPROM */
    cmd=CMD_ERASE;
    if((exitcode=uprom_action(cmd))) return(exitcode); 
  } 

  if(args_info->blank_given) {				/* blank test */
    cmd=CMD_BLANK_COPY;
    if((exitcode=uprom_action(cmd))) return(exitcode); 
  } 
  
  if(args_info->read_given) {				/* read chip */
    cmd=CMD_READ_COPY;
    if(args_info->master_socket_given) {
      cmd=CMD_READ_MASTER;
      if(ictyp==5) /* 2532 */ 
        lvprintf(VB_ERR, "***** WARNING *****\n"
        "uprom's firmware might be buggy and "
        "send you the first 2KB page twice!\n"
        "Use the copy socket instead.\n");
    }
    errcde=uprom_action(cmd);
    if(errcde) return 3;
    errcde=exitcode=save_buffer(size, offset, args_info->read_arg);
  }  
  
  if(args_info->write_given) {				/* write chip */
    if(args_info->master_socket_given) {
      lvprintf(VB_ERR, "Master socket is for reading only! Use copy socket.\n");
      return(3);
    }
    if(args_info->blank_given) sleep(1);
    errcde=load_buffer(size, offset, args_info->write_arg);
    if(errcde) return 3;
    cmd=CMD_PRG_COPY;
    errcde=uprom_action(cmd);
  } 
  
  if(args_info->compare_given) {		/* compare master == copy */
    cmd=CMD_COPY_EQU_MASTER;
    return (uprom_action(cmd));
  }  
  
  close(fd);
  return errcde;
}

void dump_all_incoming(void)
{
  int rd;
  char rcbuffer[8];
  
  printf("Dumping all incoming bytes -- press Ctrl-C to abort\n");
  for(;;) {
    rd=read(fd, rcbuffer, 1);
    printf("%02X ", rcbuffer[0] & 0xFF);
  } 
}

void uprom_init_errmsg(void) 
{
  int i;
  
  for(i=0; i<24; i++) errmsg[i][0]=0;
  strcpy(errmsg[0], "Insufficient buffer size");
  strcpy(errmsg[1], "Undefined command");
  strcpy(errmsg[2], "Wrong chip no.");
  strcpy(errmsg[3], "Invalid address");
  strcpy(errmsg[4], "Start address beyond limits");
  strcpy(errmsg[5], "EPROM not empty");
  strcpy(errmsg[6], "Unable to program");
}


int uprom_read_eprom(int cmd, int ictyp, long from, long size)
{
  int rd, errcde;
  long sum; 
  HML hml_from; HML hml_size;
  char cmdbuffer[16];
  
  lvprintf(VB_DBG, "uprom_read_eprom: ictyp=%02X from=%lX size=%lX\n", \
                                            ictyp, from, size);
  long2hml(from, &hml_from); long2hml(size, &hml_size);  
  
  cmdbuffer[0]=0;		/* Command is 8 bytes long */
  cmdbuffer[1]=8;
  cmdbuffer[2]=cmd;
  cmdbuffer[3]=ictyp;
  cmdbuffer[4]=hml_from.m;
  cmdbuffer[5]=hml_from.l;
  cmdbuffer[6]=hml_size.m;
  cmdbuffer[7]=hml_size.l;
  show_cmdbuffer(cmdbuffer, 8);
  send_cmdbuffer(cmdbuffer, 8);  

  lvprintf(VB_DBG, "Waiting for result...\n");
  do {
    rd=read(fd, buffer, 1);
  } while(!rd);

  errcde=buffer[0];
  lvprintf(VB_DBG, "Result: %Xh\n", errcde);
    
  if(errcde) {
    lvprintf(VB_ERR, "Programmer reported error %02X\n", errcde);
    return(errcde);    
  } else
  {
    lvprintf(VB_DBG, "Reading data...\n");
    sum=0;
    while(sum<size) {
      statln(sum, size);
      rd=read(fd, buffer+from+sum, size-sum);
      if(rd<0) {
        perror("reading serial data");
        break;
      }
      sum+=rd;
      lvprintf(VB_DBG, "%d bytes received, total %ld of %ld.\n", rd, sum, size);
    }
    endstatln(size, size); 
    show_buffer(from, size);
    return 0;
  } /* if(errcde) */
}

int uprom_blank_copy(int ictyp)
{
  int rd, errcde;
  char cmdbuffer[16];
  char buffer[2];
  
  lvprintf(VB_DBG, "blank_copy: ictyp=%02X\n", ictyp);
  
  cmdbuffer[0]=0;	/* command is 4 bytes long */
  cmdbuffer[1]=4;
  cmdbuffer[2]=3;	/* Blank test copy socket */
  cmdbuffer[3]=(ictyp==2816 ? 0 : ictyp);
  
  show_cmdbuffer(cmdbuffer, 4);
  send_cmdbuffer(cmdbuffer, 4);

  do {
    rd=read(fd, buffer, 1);
  } while(!rd);
  errcde=buffer[0] & 0xFF;
  lvprintf(VB_DBG, "er: %Xh\n", errcde);

  return(errcde);
}


int uprom_eeprom_erase(int ictyp)
{
  int rd, errcde, sum;
  char cmdbuffer[16];
  char buffer[4];
  
  lvprintf(VB_DBG, "eeprom_erase: ictyp=%02X\n", ictyp);
  cmdbuffer[0]=0;	/* command is 4 bytes long */
  cmdbuffer[1]=4;
  cmdbuffer[2]=5;	/* CMD_ERASE */
  cmdbuffer[3]=(ictyp==2816 ? 0 : ictyp);	
  show_cmdbuffer(cmdbuffer, 4);
  send_cmdbuffer(cmdbuffer, 4);

  sum=0;
  while(sum<1) {
    rd=read(fd, buffer+sum, 1-sum);
    sum+=rd;
  }
  errcde=buffer[0] & 0xFF;
  lvprintf(VB_DBG, "er: %02X\n", errcde);
  if(!errcde) lvprintf(VB_INFO, "EEPROM erased.\n");
  return(errcde);
}

int uprom_prg_1buffer(int ictyp, long from, long size)
{
  int rd, wr, errcde;
  HML hml_from; HML hml_size; 
  char cmdbuffer[16];
  char rcbuffer[4];
  
  lvprintf(VB_DBG, "prg_1buffer: ictyp=%02X from=%lX size=%lX\n", ictyp, from, size);
  
  long2hml(from, &hml_from); 
  long2hml(size+6, &hml_size);  
  
  cmdbuffer[0]=hml_size.m;
  cmdbuffer[1]=hml_size.l;
  cmdbuffer[2]=2; 					/* fast programming mode */
  if(uprom_prog_mode==PROG_MODE_STD) cmdbuffer[2]=4;	/* std  programming mode */
  cmdbuffer[3]=(ictyp==2816 ? 0 : ictyp);
  cmdbuffer[4]=hml_from.m;
  cmdbuffer[5]=hml_from.l;
  show_cmdbuffer(cmdbuffer, 6);
  send_cmdbuffer(cmdbuffer, 6);  
  
  wr=write(fd, buffer+from, size);		/* programm data */
  
  lvprintf(VB_DBG, "Waiting for result... (should be: 0x01: start programming)\n");
  do {
    rd=read(fd, rcbuffer, 1);
  } while(!rd);

  errcde=rcbuffer[0] & 0xFF;
  lvprintf(VB_DBG, "Result: %Xh\n", errcde);
  
  if(errcde!=1) {
    lvprintf(VB_ERR, "programming failed with msg %02Xh\n", errcde);
    return(errcde);
  } else {
    /* 01: okay, programming started... */
    lvprintf(VB_DBG, "msg 0x01: programming started\n");
  }
    
  lvprintf(VB_DBG, "Waiting for result... (should be: 0x00: programming succeded)\n");
  do {
    rd=read(fd, rcbuffer, 1);
  } while(!rd);

  errcde=rcbuffer[0] & 0xFF;
  lvprintf(VB_DBG, "Result: %Xh\n", errcde);
  
    
  return(errcde);    
}

int uprom_prg_copy(int ictyp, long from, long size)
{
  int errcde;			/* error code */
  long current_size; 		/* bytes to be written this time */
  long sum;			/* bytes already written */
  long total;			/* bytes yet to be written */
  
  lvprintf(VB_DBG, "prg_copy: ictyp=%02Xh from=%lXh size=%lXh\n", \
                                              ictyp, from, size);

  sum=0; total=size;  
  do {
    current_size=total;
    if(current_size>PROMBUFSIZE) {
      total-=PROMBUFSIZE;
      current_size=PROMBUFSIZE;
    }
    lvprintf(VB_DBG, \
    "Programming %lXh bytes starting at %lXh\n", current_size, from+sum);
    statln(sum, size);
    errcde=uprom_prg_1buffer(ictyp, from+sum, current_size);
    if(errcde) {
      statln(-1, 0);
      return errcde;
    }
    sum+=current_size;
  } while(sum<size);    
  endstatln(size, size);
  return errcde;
}

int uprom_action(int cmd)
{
  int errcde, exitcode;
  
  lvprintf(VB_DBG, "action %d: ", cmd);
  errcde=exitcode=0;
  

  switch(cmd) {
    case CMD_READ_MASTER:
    case CMD_READ_COPY:
      errcde=exitcode=uprom_read_eprom(cmd, ictyp, from, size);
      break;
      
    case CMD_BLANK_COPY:
      errcde=uprom_blank_copy(ictyp);
      if(errcde) {
        if(errcde==0x03) {
          lvprintf(VB_ERR, "Not blank.\n");
          errcde=0; exitcode=1;
        } 
      } else {
        lvprintf(VB_INFO, "Blankcheck passed.\n");
      } 
      break;
    
    case CMD_ERASE:
      errcde=exitcode=uprom_eeprom_erase(ictyp);
      break;
      
    case CMD_PRG_COPY:
      errcde=exitcode=uprom_prg_copy(ictyp, from, size);
      break;

    default:
      lvprintf(VB_ERR, "Missing command (see help/actions)\n");
      return(3);
  }
  if(errcde) errhdl(errcde);
  return(exitcode);
}

long uprom_sizefromictyp(int ictyp)
{
  switch(ictyp) {
    case 0: return 2048;
    case 1: return 4096;
    case 2816: return 2048;
    case 2: return 8192;
    case 3: return 16384;
    case 4: return 32768;
    case 5: return 4096;
  }
  return -1;
}


